package CleanTeX;
# This package prepares a LaTeX file for translation to LyX
# - Translates some local commands (e.g., {\em blah} to {\emph{blah}})
# - Prepares math mode stuff for LyX. LyX reads LaTeX math mode directly,
#      so reLyX can basically copy all math mode exactly, but LyX is a
#      bit stricter than LaTeX. E.g., translate 'x^2' -> 'x^{2}

use strict;

######
# Global variables
my $last_printstr = ""; # last string we printed


#########################   PARSER INVOCATION   ################################
sub call_parser {
# This subroutine opens the TeX parser and processes the file.
# Arg0 is the name of the input TeX file
# Arg1 is the name of the output "clean" file

    # This is the list of tokens for the parser
    # Parts of the token list were swiped from TeX.pm
    my %MyTokens = ( '{' => $Text::TeX::Tokens{'{'},
		     '}' => $Text::TeX::Tokens{'}'},
		     '$' => $Text::TeX::Tokens{'$'},
		     '$$' => $Text::TeX::Tokens{'$$'},
		     '\begin' => $Text::TeX::Tokens{'\begin'},
		     '\end' => $Text::TeX::Tokens{'\end'},
		     '\em' => $Text::TeX::Tokens{'\em'},
		     '\rm' => $Text::TeX::Tokens{'\em'},
		     '\bf' => $Text::TeX::Tokens{'\em'},
		     '\tt' => $Text::TeX::Tokens{'\em'},
		     '\sf' => $Text::TeX::Tokens{'\em'},
		     '\sc' => $Text::TeX::Tokens{'\em'},
		     '\sl' => $Text::TeX::Tokens{'\em'},
		     '\it' => $Text::TeX::Tokens{'\em'},
		     '\rmfamily' => $Text::TeX::Tokens{'\em'},
		     '\ttfamily' => $Text::TeX::Tokens{'\em'},
		     '\sffamily' => $Text::TeX::Tokens{'\em'},
		     '\mdseries' => $Text::TeX::Tokens{'\em'},
		     '\bfseries' => $Text::TeX::Tokens{'\em'},
		     '\upshape' => $Text::TeX::Tokens{'\em'},
		     '\itshape' => $Text::TeX::Tokens{'\em'},
		     '\slshape' => $Text::TeX::Tokens{'\em'},
		     '\scshape' => $Text::TeX::Tokens{'\em'},
		     # \cal is "local", although it's found in math mode
		     '\cal' => $Text::TeX::Tokens{'\em'},
		     # Unlike $TT::Tokens, we don't give ^ a 'lookahead'
		     '^' => {Type => 'report_args', count => 1},
		     '_' => {Type => 'report_args', count => 1},
		     '\frac' => {Type => 'report_args', count => 2},
		    ); 

    my ($InFileName, $OutFileName) = (shift,shift);

    warn "Cleaning TeX file... \n";
    open (OUTFILE, ">$OutFileName");

    my $file = new Text::TeX::OpenFile 
	   $InFileName,
	   'defaultact' => \&clean_tex,
	   'tokens' => \%MyTokens;

    # Now actually process the file
    $file->process;
    close OUTFILE;
    warn "Done cleaning TeX file\n";
} # end sub call_parser


#######################   MAIN TRANSLATING SUBROUTINE   ########################
# Routine called by the TeX-parser to perform token-processing.
sub clean_tex {
    my($eaten,$txt) = (shift,shift);
    my ($outstr, $type);

    # (We could make the translation tables a hash of hashes, but why bother?)
    # Translation table for TT::Begin::Group tokens
    my %begtranstbl = (
			'$' => '\(', # LyX math mode doesn't
			'$$' => '\[', # understand \$ or $$
			);

    # Translation table for TT::End::Group tokens
    my %endtranstbl = (
			   '$' => '\)',
			   '$$' => '\]',
		       );

    # Translation table for TT::Token tokens whose translations should
    #    NOT have whitespace after them! See sub translate...
    #   Note that tokens of type TT::EndLocal are always translated to '}'. So,
    #   any token defined as a local token *must* be translated to something
    #   with a '{' (e.g., '\em' -> '\emph{') or we'll have mismatched braces
    my %no_ws_transtbl = (
			'\em' => '\emph{',
			'\rm' => '\textrm{',
			'\bf' => '\textbf{',
			'\tt' => '\texttt{',
			'\sf' => '\textsf{',
			'\sc' => '\textsc{',
			'\sl' => '\textsl{',
			'\it' => '\textit{',
			'\rmfamily' => '\textrm{',
			'\ttfamily' => '\texttt{',
			'\sffamily' => '\textsf{',
			'\mdseries' => '\textmd{',
			'\bfseries' => '\textbf{',
			'\upshape' => '\textup{',
			'\itshape' => '\textit{',
			'\slshape' => '\textsl{',
			'\scshape' => '\textsc{',
			'\cal' => '\mathcal{'
			);


    # a faux "switch" statement.  sets $_ for later use in pattern
    # matching.
    $type = ref($eaten);
    $type =~ s/^Text::TeX::// or die "Non-Text::TeX object";
    my $printstr = ""; # default for undefined printstrs etc.
    SWITCH: for ($type) {
	   # Handle blank lines.
	   if (/Paragraph/) {
		  last SWITCH;
	   }

	   # Handle the end of a local font command - insert a '}'
	   if (/EndLocal/) {
	       # we could just say $printstr='}'
	       $printstr = &translate('}', \%endtranstbl);
	       last SWITCH;
	   }
	   
	   # $eaten->exact_print is undefined for previous environments
	   $outstr = $eaten->exact_print;
	   if (! defined $outstr) { # comment at end of paragraph
	       warn "Weird undefined token $eaten!" unless $eaten->[1];
	       last SWITCH;
	    }
	   
	   # Handle LaTeX tokens
	   if (/^Token$/) {
	       # HACK to remove whitespace, so that '\em ' -> '\emph{'
	       my $realtok = $eaten->print; # w/out whitespace
	       if (exists $no_ws_transtbl{$realtok}) {
		   $printstr = &translate($outstr, \%no_ws_transtbl, "no_ws");
	       } else {
	           $printstr = $outstr;
	       }
	       last SWITCH;
	   }

	   # Tokens taking arguments, like '^'
	   # ADD '{' if there isn't one before the argument!
	   if (/^BegArgsToken$/) {
	       $printstr = $outstr;
	       my $tok = $txt->lookAheadToken;
	       $printstr .= '{' unless $tok =~ /\{/;
	       last SWITCH;
	   }

	   # End of one argument, beginning of next
	   # Note that by default these tokens don't print anything
	   # ADD '}' if there isn't one after the last argument, i.e., 
	   #     if the last thing you printed *wasn't* a '}'
	   # ADD '{' if there isn't one before the next argument!
	   if (/^ArgToken$/) {
	       $printstr = $outstr; # = ''
	       $printstr .= '}' unless $last_printstr =~ /\}/;
	       my $tok = $txt->lookAheadToken;
	       $printstr .= '{' unless $tok =~ /\{/;
	       last SWITCH;
	   }

	   # End of tokens taking arguments, like '^'
	   # Note that by default these tokens don't print anything
	   # ADD '}' if there isn't one after the last argument, i.e., 
	   #     if the last thing you printed *wasn't* a '}'
	   if (/^EndArgsToken$/) {
	       $printstr = $outstr; # = ''
	       $printstr .= '}' unless $last_printstr =~ /\}/;
	       last SWITCH;
	   }
	   
	   # Handle opening groups, like '{' and '$'.
	   if (/Begin::Group$/) {
	       $printstr = &translate($outstr,\%begtranstbl);
	       last SWITCH;
	   }
	   
	   # Handle closing groups, like '}' and '$'.
	   if (/End::Group$/) {
	       $printstr = &translate($outstr, \%endtranstbl);
	       last SWITCH;
	   }

	   if (/Begin::Group::Args/ || /End::Group::Args/) {
	       $printstr = $outstr;
	       last SWITCH;
	   }
	   
	   if (/Text/) {
	       $printstr = $outstr;
	       last SWITCH;
	   }

	   # The default action - print the string.
	   $printstr = $outstr;
    } # end SWITCH:for ($type)
    
    # Actually print the string
    if (defined $printstr) { 
	print OUTFILE $printstr;
	$last_printstr = $printstr; #save for next time
    } else {warn "Undefined printstr";}

} # end sub clean_tex

####################   TRANSLATOR SUBROUTINES    ###############################
sub translate {
# Replace a string (possibly with whitespace around it) with another
# Arg0 is a string, Arg1 is a reference to a hash containing translations
# If a token not in the table is passed in, do nothing
# If Arg2 is defined AND the token is known, the remove whitespace from
#     the end of the translated token. This is a HACK to do '\em ' -> '\emph{'
# Return the string, possibly modified
    my ($tokstr, $transref) = (shift, shift);
    my $remove_ws = shift;
    my %transtable = %$transref;

    # remove whitespace from the string (since transtable doesn't have it)
    my $stripstr = $tokstr;
    $stripstr =~ s/^\s*(\S+)\s*$/$1/ or die "coudln't strip token";
    if ( exists $transtable{$stripstr} ) {
         # use \Q or \, (, $, and [ will be misinterpreted
        $tokstr =~ s/\Q$stripstr\E/$transtable{$stripstr}/;

	# remove whitespace?
	if (defined $remove_ws) {
	    $tokstr =~ s/\s*$//;
	}
    }

    return $tokstr;
}


1; # return true value to calling program
