#!/usr/bin/perl -w
#
# Main code for reLyX - the LaTeX to LyX translator
# This code usually gets called by the reLyX wrapper executable
#
# $Id: reLyXmain.pl,v 1.1.1.1 1998/04/23 16:02:34 larsbj Exp $
#

require 5.002; # Perl 5.001 doesn't work. Perl 4 REALLY doesn't work.

# Standard Perl Library modules
use strict; # must define all variables, no barewords, no hard references

# Variables from other files, like Getopt::Std and reLyX wrapper
use vars qw($opt_c $opt_d $opt_f $lyxdir);
use Getopt::Std; # read in command-line options
use File::Basename; # &basename and &dirname

# Modules containing subroutines for reLyX
# Note that @INC must include the directory that these modules are in
use Text::TeX; # TeX parser package
use MakePreamble; # package to split off LaTeX preamble & translate it to LyX
use CleanTeX; # package to clean TeX file for Lyxifying
use BasicLyX; # package to translate clean TeX to Basic LyX
use LastLyX; # package to print out LyX file once all translation is done


# variables that a user might want to change
my @suffixlist = '\.(ltx|latex|tex)'; # allowed suffixes for TeX file
my $LyXFormat = "2.15"; #What to print in \lyxformat command in .lyx file


##############################################################################
# MAIN
#

# Important variables!
my ($InFileName, $OutFileName); # current in/output file names
my $PreambleName; # name of the file we store the LaTeX preamble in
my $PathBase; # InFileName without .tex ending. For building output filenames
my @deletelist; # files to delete when we're done (if no '-d' flag)
my $doc_class; # LaTeX documentclass
my $LyXPreamble; # String in which we hold the *entire* LyX Preamble

# layoutdir is the directory where LyX finds its layouts.
# It uses $lyxdir, calculated by configure & stored in the reLyX wrapper
# The "or" below is a hack to allow running reLyX without the wrapper
my $layoutdir = ($lyxdir || "/usr/local/share/lyx") . "/layouts/";
my $LayoutFileName; # File to read LyX layouts from


#get arguments to reLyX, including TeX file name
($InFileName, $PathBase) = &getargs;

# Split the preamble off of the rest of the file
# Copy until \begin{document} into foo.preamble
# Copy rest into foo.relyx1
$PreambleName = $PathBase . ".preamble";
push @deletelist, $PreambleName; # delete foo.preamble when done
my $filenum = 1; #numbering for temporary files
$OutFileName = $PathBase . ".relyx" . $filenum;
push @deletelist, $OutFileName;
&MakePreamble::split_preamble($InFileName, $PreambleName, $OutFileName);

# Now read and translate the LaTeX preamble into LyX
# Return the document's textclass so we know which layout file(s) to read
# Also return the LyX preamble in a string, for later
($doc_class,$LyXPreamble) =
         &MakePreamble::translate_preamble($PreambleName, $LyXFormat);

# Clean the TeX file (not including its preamble)
$InFileName = $OutFileName;
$OutFileName = $PathBase . ".relyx" . ++$filenum; # foo.relyx1 -> foo.relyx2
push @deletelist, $OutFileName;
&CleanTeX::call_parser($InFileName, $OutFileName);

# Now convert basic constructs in the cleaned TeX file to LyX constructs
$LayoutFileName = $layoutdir . $doc_class . ".layout";
$InFileName = $OutFileName;
$OutFileName = $PathBase . ".relyx" . ++$filenum; # foo.relyx2 -> foo.relyx3
push @deletelist, $OutFileName;
&BasicLyX::call_parser($InFileName, $OutFileName, $LayoutFileName);

# Someday, there will be calls to class-specific translation routines here
# Or will there?...

# Finally, print out the actual LyX file including the preamble
$InFileName = $OutFileName;
$OutFileName = $PathBase . ".lyx";
&LastLyX::last_lyx($InFileName, $OutFileName, $LyXPreamble);

unless ($opt_d) {
    warn "Deleting @deletelist\n";
    unlink @deletelist;
}

warn "Finished successfully!\n";

##################### SUBROUTINES  ###########################################
# Gets the script's arguments.
# Returns the input file name, as well as the 'pathbase', which
#    is the input file name (including path) without the '.tex' ending.
#    It will be used for building temporary and output file names
sub getargs {
    # set $opt_f if '-f' on command line, -r takes an argument
    getopts('c:dfr:') or die "Illegal option! man reLyX for details\n";
    my $InFileName = $ARGV[0];
    die "No filename specified.\n" unless defined $InFileName;
    die "Cannot find input file $InFileName.\n"   unless -e $InFileName;

    # Get file names, set up for making different temporary files
    # fileparse_set_fstype("MSDOS") for DOS support!
    my ($basename, $path, $suffix) = fileparse($InFileName, @suffixlist);
    die "invalid file suffix" unless $suffix;
    $path .= '/' unless $path =~ /\/$/; # fix BUG in perl5.002 fileparse!
    my $PathBase = $path . $basename; # all but suffix

    # Check for files that already exist
    my $pname = $PathBase . ".preamble";
    my $lname = $PathBase . ".lyx";
    if ($opt_f) {
	warn "Will overwrite Preamble file $pname\n" if -e $pname;
	warn "Will overwrite LyX file $lname\n" if -e $lname;
    } else {
	die "Preamble file $pname already exists\n" if -e $pname;
	die "LyX file $lname already exists\n" if -e $lname;
    }

    return ($InFileName, $PathBase);
}


1; # return "true" to the wrapper
